--------------------------------------------------------------------------------
-- From the balanced view to the right-biased view
--------------------------------------------------------------------------------

{-# LANGUAGE DataKinds              #-}
{-# LANGUAGE PolyKinds              #-}
{-# LANGUAGE TypeFamilies           #-}
{-# LANGUAGE GADTs                  #-}
{-# LANGUAGE FlexibleContexts       #-}
{-# LANGUAGE FlexibleInstances      #-}
{-# LANGUAGE ScopedTypeVariables    #-}
{-# LANGUAGE TypeOperators          #-}
{-# LANGUAGE UndecidableInstances   #-}

module Conversion.S2RS where

import Library.Structured
import Auxiliary.TypeLevel


--------------------------------------------------------------------------------
-- Type-level conversion
--------------------------------------------------------------------------------

-- Datatypes
type family S2RS (a :: UData k) :: UData k
type instance S2RS ('UData md t) = 'UData md (S2RSConR t)

-- Constructors
type family S2RSConR (a :: Tree (UCon k)) :: Tree (UCon k)
type instance S2RSConR  Empty              = Empty
type instance S2RSConR (Leaf ('UCon mc t)) = Leaf ('UCon mc (S2RSSelR t))
type instance S2RSConR (Bin (Bin x1 x2) y) = S2RSConR (RotateR (Bin (Bin x1 x2) y))
type instance S2RSConR (Bin (Leaf x)    y) = Bin (S2RSConR (Leaf x)) (S2RSConR y)

type family RotateR (a :: Tree k) :: Tree k
type instance RotateR (Bin (Bin x1 x2) y) = Bin x1 (Bin x2 y)
type instance RotateR (Bin (Leaf x)    y) = Bin (Leaf x) y
type instance RotateR (Bin Empty       y) = Bin Empty (RotateR y) -- can't happen, I think

-- Selectors
type family S2RSSelR (a :: Tree (USel k)) :: Tree (USel k)
type instance S2RSSelR  Empty              = Empty
type instance S2RSSelR (Leaf t)            = Leaf t -- done
type instance S2RSSelR (Bin (Bin x1 x2) y) = S2RSSelR (RotateR (Bin (Bin x1 x2) y))
type instance S2RSSelR (Bin (Leaf x)    y) = Bin (Leaf x) (S2RSSelR y)

--------------------------------------------------------------------------------
-- Value-level conversion
--------------------------------------------------------------------------------

-- Datatypes
class Convert (a :: UData k) where
  s2rs :: In a p -> In (S2RS a) p
  rs2s :: In (S2RS a) p -> In a p

instance (ConvertConR t) => Convert ('UData md t) where
  s2rs (D1 x) = D1 (s2rsConR x)
  rs2s (D1 x) = D1 (rs2sConR x)

-- Constructors: recursively rotate
class ConvertConR (a :: Tree (UCon k)) where
  s2rsConR :: In a p -> In (S2RSConR a) p
  rs2sConR :: In (S2RSConR a) p -> In a p

instance (ConvertSelR t) => ConvertConR (Leaf ('UCon mc t)) where
  s2rsConR (C1 x) = C1 (s2rsSelR x)
  rs2sConR (C1 x) = C1 (rs2sSelR x) 

-- Rotate right until we can't rotate anymore
instance ( ConvertConR (RotateR    (Bin (Bin x1 x2) y))
         ,              ConvertCon (Bin (Bin x1 x2) y))
        => ConvertConR (Bin (Bin x1 x2) y) where
  s2rsConR = s2rsConR . s2rsCon
  rs2sConR = rs2sCon . rs2sConR

instance (ConvertConR (Leaf x), ConvertConR y)
        => ConvertConR (Bin (Leaf x) y) where
  s2rsConR (L1 x) = L1 (s2rsConR x)
  s2rsConR (R1 y) = R1 (s2rsConR y)
  rs2sConR (L1 x) = L1 (rs2sConR x)
  rs2sConR (R1 y) = R1 (rs2sConR y)

-- Constructors: rotate once
class ConvertCon (a :: Tree (UCon k)) where
  s2rsCon :: In a p -> In (RotateR a) p
  rs2sCon :: In (RotateR a) p -> In a p

instance ConvertCon (Bin (Leaf x) y) where
  s2rsCon = id
  rs2sCon = id

instance ConvertCon (Bin (Bin x1 x2) y) where
  s2rsCon (L1 (L1 x1)) = L1 x1
  s2rsCon (L1 (R1 x2)) = R1 (L1 x2)
  s2rsCon (R1 y)       = R1 (R1 y)

  rs2sCon (L1 x1)      = L1 (L1 x1)
  rs2sCon (R1 (L1 x2)) = L1 (R1 x2)
  rs2sCon (R1 (R1 y))  = R1 y

-- Selectors: recursively rotate
class ConvertSelR (a :: Tree (USel k)) where
  s2rsSelR :: In a p -> In (S2RSSelR a) p
  rs2sSelR :: In (S2RSSelR a) p -> In a p

instance ConvertSelR Empty where
  s2rsSelR U1 = U1
  rs2sSelR U1 = U1

instance ConvertSelR (Leaf t) where
  s2rsSelR x = x
  rs2sSelR x = x

instance ( ConvertSelR (RotateR    (Bin (Bin x1 x2) y))
         ,              ConvertSel (Bin (Bin x1 x2) y))
        => ConvertSelR (Bin (Bin x1 x2) y) where
  s2rsSelR = s2rsSelR . s2rsSel
  rs2sSelR = rs2sSel . rs2sSelR

instance (ConvertSelR y) => ConvertSelR (Bin (Leaf x) y) where
  s2rsSelR (x :*: y) = x :*: s2rsSelR y
  rs2sSelR (x :*: y) = x :*: rs2sSelR y


-- Selectors: rotate once
class ConvertSel (a :: Tree (USel k)) where
  s2rsSel :: In a p -> In (RotateR a) p
  rs2sSel :: In (RotateR a) p -> In a p

instance ConvertSel (Bin (Leaf x) y) where
  s2rsSel = id
  rs2sSel = id

instance ConvertSel (Bin (Bin x1 x2) y) where
  s2rsSel ((x1 :*: x2) :*: y) = x1 :*: (x2 :*: y)
  rs2sSel (x1 :*: (x2 :*: y)) = (x1 :*: x2) :*: y
